/*  Copyright (C) 2009 Thilo Molitor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//TODO: -U *:r/v:* muss noch untersttzt werden
//TODO: Alle argumente von avrdude mssen zumindest ignoriert werden
//DONE but untested: bei schecksum error block neu bertragen
//TODO: Device puffer mit 65536 zeichen zu klein gewhlt
//TODO: read and write eeprom funktioniert noch nicht


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */
#define FALSE 0
#define TRUE 1


//Bootloaderbefehle von protokoll v0.1
#define CMD_DUMMY		255
#define CMD_CONNECT		4
#define CMD_ERASE_FLASH 	5
#define CMD_WRITE_FLASH 	6
#define CMD_READ_FLASH		7
#define CMD_FLASH_BLOCKSIZE 	8
#define CMD_JUMP_PROGRAM 	9
//Bootloaderbefehle von protokoll v0.2
#define CMD_READ_LFUSE 		10
#define CMD_READ_HFUSE 		11
//Bootloaderbefehle von protokoll v0.3
#define CMD_READ_EEPROM		12
#define CMD_WRITE_EEPROM	13
#define CMD_EEPROM_BLOCKSIZE 	14

//messages von protokoll v0.1
#define MSG_END			255
#define MSG_OVERWRITE_ERROR	4
#define MSG_CHKSUM_ERROR	5
//keine neuen messages von protokoll v0.2
//keine neuen messages von protokoll v0.3

//definierte befehle und antworten
#define CMD_EMPTY		1
static unsigned char commands[][64]={
		{CMD_EMPTY},
		{CMD_DUMMY, CMD_CONNECT, CMD_ERASE_FLASH, CMD_WRITE_FLASH, CMD_READ_FLASH, CMD_FLASH_BLOCKSIZE, CMD_JUMP_PROGRAM, CMD_EMPTY},
		{CMD_DUMMY, CMD_CONNECT, CMD_ERASE_FLASH, CMD_WRITE_FLASH, CMD_READ_FLASH, CMD_FLASH_BLOCKSIZE, CMD_JUMP_PROGRAM, CMD_READ_LFUSE, CMD_READ_HFUSE, CMD_EMPTY},
		{CMD_DUMMY, CMD_CONNECT, CMD_ERASE_FLASH, CMD_WRITE_FLASH, CMD_READ_FLASH, CMD_FLASH_BLOCKSIZE, CMD_JUMP_PROGRAM, CMD_READ_LFUSE, CMD_READ_HFUSE, CMD_READ_EEPROM, CMD_WRITE_EEPROM, CMD_EEPROM_BLOCKSIZE, CMD_EMPTY}
};


struct termios oldtio,newtio;

int jump=1, erase_flash=0;
char MODEMDEVICE[512];
int baud=0;
char file[8192];
char action[8][8192];
int action_counter=0;
int legacy_with_timeout=0;
struct timeval timeout;
int rate[]={4800, 9600, 38400};
//int rate[]={600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200};
int rate_num=3;
int rate_cur=0;
int proto=0;
int safemode=0, safeblocks=0;


//rechnet die protokollversion in die eintragsnummer vom commands array um
int protokoll_supported(int protokoll)
{
	switch(protokoll)
	{
		case 1: return 1;		//v0.1
		case 2: return 2;		//v0.2
		case 3: return 3;		//v0.3
	}
	return 0;				//unbekannt
}

int command_supported(unsigned char command, int protokoll)
{
	int i;
	for(i=0;commands[protokoll][i]!=CMD_EMPTY;i++)
		if(commands[protokoll][i]==command)
			return 1;
	return 0;
}

speed_t get_rate(int baud)
{
	/** This switch statement features some code from minicom 2.0.0, src/sysdep1.c and cutecom 0.14.0*/
	switch (baud)
	{
		case 600:	return B600;
		case 1200:	return B1200;
		case 2400:	return B2400;
		case 4800:	return B4800;
		case 9600:	return B9600;
//		case 14400:	return B14400;
		case 19200:	return B19200;
//		case 28800:	return B28800;
		case 38400:	return B38400;
//		case 56000:	return B56000;
		case 57600:	return B57600;
		case 115200:	return B115200;
		case 230400:	return B230400;
		case 460800:	return B460800;
		case 576000:	return B576000;
		case 921600:	return B921600;
//		case 128000:	return B128000;
//		case 256000:	return B256000;
		default:
			if(baud==0)
				printf("WARNING: Baudrate not specified, scanning for baudrate\n");
			else
				printf("WARNING: Baudrate %d unknown, defaulting to baudrate 4800\n", baud);
			return B4800;
	}
}

/** This function features some code from minicom 2.0.0, src/sysdep1.c and cutecom 0.14.0*/
int init_terminal(char *MODEMDEVICE, speed_t _baud)
{
	int fd,c, res;
	
	fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY ); 
	if (fd <0) 
	{
		printf("ERROR\n");
		perror(MODEMDEVICE); exit(-1); 
	}
	
	tcgetattr(fd,&oldtio);	/* save current port settings */
	tcgetattr(fd,&newtio);	/* save current port settings */
	
	cfsetospeed(&newtio, _baud);
	cfsetispeed(&newtio, _baud);
	newtio.c_cflag = (newtio.c_cflag & ~CSIZE) | CS8;
	newtio.c_cflag |= CLOCAL | CREAD;
	newtio.c_cflag &= ~(PARENB | PARODD);
	newtio.c_cflag &= ~CRTSCTS;
	newtio.c_cflag &= ~CSTOPB;
	newtio.c_iflag &= ~(IXON|IXOFF|IXANY);
	newtio.c_lflag=0;
	newtio.c_oflag=0;
	newtio.c_cc[VTIME]=1;
	newtio.c_cc[VMIN]=1;

	tcflush(fd, TCIFLUSH);
	tcsetattr(fd,TCSANOW,&newtio);
	
	return fd;
}
void deinit_terminal(int fd)
{
	tcsetattr(fd,TCSANOW,&oldtio);
	close(fd);
	return;
}

void uart_putc(int term_fd, unsigned char byte)
{
	if(write(term_fd, &byte, 1)==-1)
	{
		fprintf(stderr, "\nERROR!!! - Cannot write to serial port...\n");
		exit(3);
	}
	fflush(NULL);		//flush all output streams (including this one)
	return;
}

int uart_get_msg(int term_fd, unsigned char *inbuffer)
{
	unsigned char *inbuffer_p;
	fd_set readfd;
	struct timeval old_timeout=timeout;
	
	FD_ZERO(&readfd);
	inbuffer_p=inbuffer;
	do
	{
		FD_ZERO(&readfd);
		if(legacy_with_timeout)
		{
			FD_SET(term_fd, &readfd);
			select(term_fd+1, &readfd, NULL, NULL, &timeout);
		}
		if(FD_ISSET(term_fd, &readfd) || !legacy_with_timeout)
		{
			if(read(term_fd, inbuffer_p++, 1)==-1)
			{
				fprintf(stderr, "\nERROR!!! - Read error on serial port...\n");
				exit(3);
			}
		}
		if(!FD_ISSET(term_fd, &readfd) && legacy_with_timeout)
		{
			timeout=old_timeout;
			return 1;		//timeout aufgetreten
		}
	} while(*(inbuffer_p-1)!=MSG_END);
	*inbuffer_p=0;
	timeout=old_timeout;
	return 0;
}

void display_usage(char **argv)
{
	printf("Usage: %s [options]\nOptions:\n",argv[0]);
		printf("\
  -P <port>                  Specify connection port.\n\
  -b <baudrate>              Override RS-232 baud rate.\n\
  -S [msec:blocksize]        Use slowmode (safemode) for communication.\n\
                             Wait msec milliseconds between blocksize bytes\n\
                             of the same data block - default 50ms each 32 bytes.\n\
  -e                         Perform a chip erase.\n\
  -J                         Do not jump to application code after operation.\n\
  -U <memtype>:r|w|v:<filename>[:format]\n\
                             Memory operation specification.\n\
                             Multiple -U options are allowed, each request\n\
                             is performed in the order specified.\n\
                             The format string is ignored due to avrdude compatibility.\n\
  -c <programmer>            Specify programmer type. (Ignored due to avrdude\n\
                             compatibility.\n\
  -p <partno>                Required. Specify AVR device. (Ignored due to avrdude\n\
                             compatibility.\n\
  -?                         Display this usage.\n\n"\
		      );
		
		return;
}

int parse_commandline(int argc, char **argv)
{
	int counter;
	
	for(counter=1; counter<argc; counter++)
	{
		if(argv[counter][0]=='-')
		{
			switch(argv[counter][1]) 
			{
				case 'S':
						safemode=50;
						safeblocks=16;
						if(++counter>=argc)
							break;
						if(argv[counter][0]=='-')
							break;
						sscanf(argv[counter], "%d:%d", &safemode, &safeblocks);
						safemode=!safemode ? 50 : safemode;
						safeblocks=!safeblocks ? 16 : safeblocks;
						break;
				case 'e':	erase_flash=1; break;
				case 'J':	jump=0; break;
				case 'P':	
						if(++counter>=argc)
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						if(argv[counter][0]=='-')
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						strcpy(MODEMDEVICE, argv[counter]); 
						break;
				case 'b':
						if(++counter>=argc)
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						if(argv[counter][0]=='-')
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						sscanf(argv[counter], "%d", &baud);
						break;
				case 'U':	
						if(++counter>=argc)
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						if(argv[counter][0]=='-')
						{
							printf("ERROR\n");
							display_usage(argv); exit(-1);
						}
						strcpy(action[action_counter++], argv[counter]);
						break;
				//avrdude compatibility commands are ignored (TODO: alle commands einfgen)
				case 'p':	
						if(argv[++counter][0]=='-')
							counter--;
						break;
				case 'c':	
						if(argv[++counter][0]=='-')
							counter--;
						break;
				case '?':	printf("OK\n"); display_usage(argv); exit(0); break;
				default:	
						printf("ERROR\n");
						fprintf(stderr, "Command '%s' not supported (yet)...\n", argv[counter]);
						printf("\n");
						display_usage(argv);
						exit(1);
						break;
			}
		}
		else
		{
			printf("ERROR\n");
			fprintf(stderr, "Expected '-%s' not '%s'...\n", argv[counter], argv[counter]);
			exit(1);
		}
	}
	return 0;
}

int read_ihex_file(FILE *fihex, int *device)
{
	unsigned char inbuffer[1024];
	unsigned int close=0, size, temp, adress, type, pos;
	int c, counter;

	while((fgets(inbuffer, 1023, fihex)!=NULL)&&(!close))
	{
		if(inbuffer[0]!=':')
		{
			fprintf(stderr, "\nERROR!!! - no ihex format (startzeichen!=':')\n");
			return 2;
		}
		if(strlen(inbuffer)<9)
		{
			//Datei unerwartet zuende
			fprintf(stderr, "\nERROR!!! - corrupt ihex format (unexpected EOF)\n");
			return 2;
		}
		sscanf(inbuffer+1, "%2x", &size);	//size
		sscanf(inbuffer+3, "%2x", &temp);	//high von adress
		sscanf(inbuffer+5, "%2x", &adress);	//low von adress
		sscanf(inbuffer+7, "%2x", &type);	//type
		adress+=(temp<<8);			//high von adress addieren
		if(strlen(inbuffer)!=13+size*2)
		{
			//Datei unerwartet zuende
			fprintf(stderr, "\nERROR!!! - corrupt ihex format (unexpected EOF)\n");
			return 2;
		}
		if(type==1)				//termination record?
			close=1;
		else
		{
			pos=adress;						//position im datenpuffer
			for(counter=0;counter<size*2;counter+=2)
			{
				sscanf(inbuffer+9+counter, "%2x", &c);		//byte einlesen
				device[pos++]=c;				//und in devicepuffer speichern
			}
		}
	}
	return pos;
}

int do_erase_flash(int term_fd)
{
	unsigned char tempchar[512];
	
	if(!command_supported(CMD_ERASE_FLASH, proto))
	{
		printf("Error: Command CMD_ERASE_FLASH not supported..\n");
		return 1;
	}
	
	printf("Erasing flash...");
	fflush(stdout);
	uart_putc(term_fd, CMD_ERASE_FLASH);		//Command erase
	uart_get_msg(term_fd, tempchar);
	if(tempchar[0]!=MSG_END)
	{
		printf("ERROR\n");
		fprintf(stderr, "Error erasing flash...\n");
		exit(80);
	}
	printf("OK\n");
	return 0;
}

int do_write_flash(int term_fd, char *file)
{
	FILE *fihex;
	unsigned char tempchar[1024];
	int device[65536];
	unsigned int flashsize, size, adress, pos;
	int chksum, counter;
	
	if(!command_supported(CMD_WRITE_FLASH, proto))
	{
		printf("Error: Command CMD_WRITE_FLASH not supported..\n");
		return 1;
	}
	if(!command_supported(CMD_FLASH_BLOCKSIZE, proto))
	{
		printf("Error: Command CMD_FLASH_BLOCKSIZE not supported..\n");
		return 1;
	}
	
	//Datei ffnen
	printf("Opening file '%s'...", file);
	fflush(stdout);
	if((fihex=fopen(file,"rt"))==NULL)
	{
		printf("ERROR\n");
		fprintf(stderr, "ERROR opening file %s...\n", file);
		//Fehler beim ffnen
		return 1;
	}
	printf("OK\n");
	printf("Reading file '%s'...", file);
	fflush(stdout);
	
	//devicepuffer leeren
	for(counter=0;counter<65536;counter++)
		device[counter]=-1;
	//Records einlesen...
	pos=read_ihex_file(fihex, device);
	fclose(fihex);
	printf("OK - read %d bytes\n", pos);
	fflush(stdout);
	
	//Recordgre ermitteln (ein Flash Block)
	uart_putc(term_fd, CMD_FLASH_BLOCKSIZE);
	uart_get_msg(term_fd, tempchar);
	flashsize=(tempchar[0]<<8)|(tempchar[1]);
	pos=0;
	while(pos<65536)
	{
		if(device[pos]!=-1)
		{
			adress=pos;								//adress festsetzen
			for(pos=adress;(pos<(adress+flashsize))&&(device[pos]!=-1);pos++)
				;
			size=pos-adress;							//size festlegen
			printf("\15Writing flash at adress=0x%04x, size=0x%02x...", adress, size);
			fflush(stdout);
			
			//Block schreiben
			//***************
			uart_putc(term_fd, CMD_WRITE_FLASH);		//Command write
			uart_putc(term_fd, adress>>8);			//page
			uart_putc(term_fd, adress&0xff);
			uart_putc(term_fd, size>>8);			//size
			uart_putc(term_fd, size&0xff);
			
			chksum=(adress>>8)%256;				//checksum auch ber commandbytes bilden
			chksum=(chksum+(adress&0xff))%256;
			chksum=(chksum+(size>>8))%256;
			chksum=(chksum+(size&0xff))%256;
			
			if(safemode)
				usleep(safemode*1000);			//sleep some time
			for(pos=adress;pos<(adress+size);pos++)		//daten senden
			{
				uart_putc(term_fd, device[pos]);
				chksum=(chksum+device[pos])%256;
				if(safemode && ((pos-adress+1)%safeblocks)==0)
					usleep(safemode*1000);		//sleep some time
			}
			uart_putc(term_fd, chksum);			//checksum senden
			
			legacy_with_timeout=1;
			if(uart_get_msg(term_fd, tempchar))
			{
				printf("TIMEOUT\nRetrying...\n");
				timeout.tv_sec=0;		//0 sekunden...
				timeout.tv_usec=250000;		//...und 250 ms timeout
				while(uart_get_msg(term_fd, tempchar) || protokoll_supported(tempchar[0]<<8 | tempchar[1])!=proto)
					uart_putc(term_fd, CMD_CONNECT);	//dummy message schreiben bis device wieder reagiert
				timeout.tv_sec=2;		//2 sekunden...
				timeout.tv_usec=0;		//...und 0 ms timeout
				pos=adress;			//gleichen block nochmal bertragen
			}
			else if(tempchar[0]==MSG_CHKSUM_ERROR)
			{
				printf("CHECKSUM_ERROR (checksum should be %u)\nRetrying...\n", chksum);
				pos=adress;		//gleichen block nochmal bertragen
			}
			else if(tempchar[0]==MSG_OVERWRITE_ERROR)
			{
				printf("\nOVERWRITE_ERROR...\n", tempchar[0], chksum);
				legacy_with_timeout=0;
				return 3;
			}
			legacy_with_timeout=0;
		}
		else
			pos++;
	}
	printf("\n");
	return 0;
}

int do_read_flash(int term_fd, char *file)
{
	FILE *fihex;
	unsigned char outbuffer[1024], *outbuffer_p, tempchar[1024];
	unsigned int size, adress, type, temp;
	int c, chksum, pos, counter;
	
	if(!command_supported(CMD_READ_FLASH, proto))
	{
		printf("Error: Command CMD_READ_FLASH not supported..\n");
		return 1;
	}
	
	//Datei ffnen
	printf("Opening %s...", file);
	fflush(stdout);
	if((fihex=fopen(file,"wb"))==NULL)
	{
		printf("ERROR\n");
		fprintf(stderr, "ERROR opening file %s...\n", file);
		//Fehler beim ffnen
		return 1;
	}
	printf("OK\n");
	
	//Recordgre ermitteln (ein Flash Block)
	uart_putc(term_fd, CMD_FLASH_BLOCKSIZE);
	uart_get_msg(term_fd, tempchar);
	size=(tempchar[0]<<8)|(tempchar[1]);
	
	
	//neuen Record beginnen...
	for(pos=0;pos<1024;pos++)		//Clear outbuffer
		outbuffer[pos]='\0';
	pos=0;
	outbuffer[pos++]=':';			//Record startzeichen in puffer schreiben
	sprintf(tempchar, "%02X", size);	//Gre in record speichern
	outbuffer[pos++]=tempchar[0];
	outbuffer[pos++]=tempchar[1];
	sprintf(tempchar, "%04X", adress);	//adress in record speichern
	outbuffer[pos++]=tempchar[0];
	outbuffer[pos++]=tempchar[1];
	outbuffer[pos++]=tempchar[2];
	outbuffer[pos++]=tempchar[3];
	outbuffer[pos++]='0';			//Type ist 00
	outbuffer[pos++]='0';
	
	//Blochbertragung anfordern
	uart_putc(term_fd, CMD_READ_FLASH);	//command senden
	uart_putc(term_fd, adress>>8);		//page
	uart_putc(term_fd, adress&0xff);
	uart_putc(term_fd, size>>8);		//size
	uart_putc(term_fd, size&0xff);
	
	//Daten empfangen
	chksum=0;
	for(counter=0;counter<size;counter++)
		if(read(term_fd, tempchar, 1)==-1)
		{
			fprintf(stderr, "\nERROR!!! - Read timeout on serial port (char %d)...\n", counter);
			exit(3);
		}
		else
		{
			sprintf(&outbuffer[pos++], "%02X", tempchar[0]);		//zeichen in hex umwandeln und in puffer speichern
			pos++;
			chksum=(chksum+tempchar[0])%256;
		}
	uart_get_msg(term_fd, tempchar);	//checksum und alles ok message einlesen
	if((tempchar[1]!=MSG_END) | (tempchar[0]!=chksum))
	{
		printf("ERROR - Block nicht sauber bertragen, breche ab...\n");
		return 2;
	}
	fprintf(fihex, "%s00\n", outbuffer);	//zeile in datei speichern
	
	return 0;
}

int do_write_eeprom(int term_fd, char *file)
{
	FILE *fihex;
	unsigned char tempchar[1024];
	int device[65536];
	unsigned int flashsize, size, adress, pos;
	int chksum, counter;
	
	if(!command_supported(CMD_WRITE_EEPROM, proto))
	{
		printf("Error: Command CMD_WRITE_EEPROM not supported..\n");
		return 1;
	}
	if(!command_supported(CMD_EEPROM_BLOCKSIZE, proto))
	{
		printf("Error: Command CMD_EEPROM_BLOCKSIZE not supported..\n");
		return 1;
	}
	
	//Datei ffnen
	printf("Opening file '%s'...", file);
	fflush(stdout);
	if((fihex=fopen(file,"rt"))==NULL)
	{
		printf("ERROR\n");
		fprintf(stderr, "ERROR opening file %s...\n", file);
		//Fehler beim ffnen
		return 1;
	}
	printf("OK\n");
	printf("Reading file '%s'...", file);
	fflush(stdout);
	
	//devicepuffer leeren
	for(counter=0;counter<65536;counter++)
		device[counter]=-1;
	//Records einlesen...
	pos=read_ihex_file(fihex, device);
	fclose(fihex);
	printf("OK - read %d bytes\n", pos);
	fflush(stdout);
	
	//Recordgre ermitteln (ein eeprom block)
	uart_putc(term_fd, CMD_EEPROM_BLOCKSIZE);
	uart_get_msg(term_fd, tempchar);
	flashsize=(tempchar[0]<<8)|(tempchar[1]);
	pos=0;
	while(pos<65536)
	{
		if(device[pos]!=-1)
		{
			adress=pos;								//adress festsetzen
			for(pos=adress;(pos<(adress+flashsize))&&(device[pos]!=-1);pos++)
				;
			size=pos-adress;							//size festlegen
			printf("\15Writing eeprom at adress=0x%04x, size=0x%02x...", adress, size);
			fflush(stdout);
			
			//Block schreiben
			//***************
			uart_putc(term_fd, CMD_WRITE_EEPROM);		//Command write
			uart_putc(term_fd, adress>>8);			//page
			uart_putc(term_fd, adress&0xff);
			uart_putc(term_fd, size>>8);			//size
			uart_putc(term_fd, size&0xff);
			
			chksum=(adress>>8)%256;				//checksum auch ber commandbytes bilden
			chksum=(chksum+(adress&0xff))%256;
			chksum=(chksum+(size>>8))%256;
			chksum=(chksum+(size&0xff))%256;
			
			for(pos=adress;pos<(adress+size);pos++)		//daten senden
			{
				uart_putc(term_fd, device[pos]);
				chksum=(chksum+device[pos])%256;
				if(safemode && ((pos-adress+1)%safeblocks)==0)
					usleep(safemode*1000);		//sleep some time
			}
			uart_putc(term_fd, chksum);			//checksum senden
			
			legacy_with_timeout=1;
			if(uart_get_msg(term_fd, tempchar))
			{
				printf("TIMEOUT\nRetrying...\n");
				pos=adress;		//gleichen block nochmal bertragen
			}
			legacy_with_timeout=0;
			if(tempchar[0]==MSG_CHKSUM_ERROR)
			{
				printf("CHECKSUM_ERROR (checksum should be %u)\nRetrying...\n", chksum);
				pos=adress;		//gleichen block nochmal bertragen
			}
			else if(tempchar[0]==MSG_OVERWRITE_ERROR)
			{
				printf("\nOVERWRITE_ERROR...\n", tempchar[0], chksum);
				return 3;
			}
		}
		else
			pos++;
	}
	printf("\n");
	return 0;
}

int do_read_eeprom(int term_fd, char *file)
{
	FILE *fihex;
	unsigned char outbuffer[1024], *outbuffer_p, tempchar[1024];
	unsigned int size, adress, type, temp;
	int c, chksum, pos, counter;
	
	if(!command_supported(CMD_READ_EEPROM, proto))
	{
		printf("Error: Command CMD_READ_EEPROM not supported..\n");
		return 1;
	}
	if(!command_supported(CMD_EEPROM_BLOCKSIZE, proto))
	{
		printf("Error: Command CMD_EEPROM_BLOCKSIZE not supported..\n");
		return 1;
	}
	
	//Datei ffnen
	printf("Opening %s...", file);
	fflush(stdout);
	if((fihex=fopen(file,"wb"))==NULL)
	{
		printf("ERROR\n");
		fprintf(stderr, "ERROR opening file %s...\n", file);
		//Fehler beim ffnen
		return 1;
	}
	printf("OK\n");
	
	//Recordgre ermitteln (EEPROM gre)
	uart_putc(term_fd, CMD_EEPROM_BLOCKSIZE);
	uart_get_msg(term_fd, tempchar);
	size=(tempchar[0]<<8)|(tempchar[1]);
	
	
	//neuen Record beginnen...
	for(pos=0;pos<1024;pos++)		//Clear outbuffer
		outbuffer[pos]='\0';
	pos=0;
	outbuffer[pos++]=':';			//Record startzeichen in puffer schreiben
	sprintf(tempchar, "%02X", size);	//Gre in record speichern
	outbuffer[pos++]=tempchar[0];
	outbuffer[pos++]=tempchar[1];
	sprintf(tempchar, "%04X", adress);	//adress in record speichern
	outbuffer[pos++]=tempchar[0];
	outbuffer[pos++]=tempchar[1];
	outbuffer[pos++]=tempchar[2];
	outbuffer[pos++]=tempchar[3];
	outbuffer[pos++]='0';			//Type ist 00
	outbuffer[pos++]='0';
	
	//Blochbertragung anfordern
	uart_putc(term_fd, CMD_READ_EEPROM);	//command senden
	uart_putc(term_fd, adress>>8);		//page
	uart_putc(term_fd, adress&0xff);
	uart_putc(term_fd, size>>8);		//size
	uart_putc(term_fd, size&0xff);
	
	//Daten empfangen
	chksum=0;
	for(counter=0;counter<size;counter++)
		if(read(term_fd, tempchar, 1)==-1)
		{
			fprintf(stderr, "\nERROR!!! - Read timeout on serial port (char %d)...\n", counter);
			exit(3);
		}
		else
		{
			sprintf(&outbuffer[pos++], "%02X", tempchar[0]);		//zeichen in hex umwandeln und in puffer speichern
			pos++;
			chksum=(chksum+tempchar[0])%256;
		}
	uart_get_msg(term_fd, tempchar);	//checksum und alles ok message einlesen
	if((tempchar[1]!=MSG_END) | (tempchar[0]!=chksum))
	{
		printf("ERROR - Block nicht sauber bertragen, breche ab...\n");
		return 2;
	}
	fprintf(fihex, "%s00\n", outbuffer);	//zeile in datei speichern
	
	return 0;
}

int do_read_lfuse(int term_fd)
{
	unsigned char tempchar[1024];
	
	if(!command_supported(CMD_READ_LFUSE, proto))
	{
		printf("Error: Command CMD_READ_LFUSE not supported..\n");
		return 1;
	}
	
	uart_putc(term_fd, CMD_READ_LFUSE);
	uart_get_msg(term_fd, tempchar);	//fuse und alles ok message einlesen
	printf("LFUSE is 0x%02X...\n", tempchar[0]);
	return 0;
}

int do_read_hfuse(int term_fd)
{
	unsigned char tempchar[1024];
	
	if(!command_supported(CMD_READ_HFUSE, proto))
	{
		printf("Error: Command CMD_READ_HFUSE not supported..\n");
		return 1;
	}
	
	uart_putc(term_fd, CMD_READ_HFUSE);
	uart_get_msg(term_fd, tempchar);	//fuse und alles ok message einlesen
	printf("HFUSE is 0x%02X...\n", tempchar[0]);
	return 0;
}

int main(int argc, char **argv)
{
	int term_fd;
	unsigned char tempchar[1024];
	int counter;
	speed_t _baud;
	int retval;
	
	printf("Simple programmer for my avr-bootloader (v0.4)...\n\n");
	if(argc<2)
	{
		display_usage(argv);
		exit(0);
	}
		
	printf("Parsing commandline...");
	parse_commandline(argc, argv);
	printf("OK\n");
	
	if(strlen(MODEMDEVICE)==0)
	{
		fprintf(stderr, "Error: No -P option specified, cannot open terminal...\n");
		display_usage(argv);
		exit(-1);
	}
	
	if(baud)
		printf("Setting up terminal %s (%d baud)...", MODEMDEVICE, baud);
	else
		printf("Setting up terminal %s...", MODEMDEVICE);
	_baud=get_rate(baud);
	term_fd = init_terminal(MODEMDEVICE, _baud);
	printf("OK\n");
	
	if(safemode)
		printf("Slowmode on - waiting %dms each %d bytes of data...\n", safemode, safeblocks);
	else
		printf("Slowmode off...\n");
	
	printf("Writing connection request to bootloader...%s", baud ? "" : "\n"); fflush(stdout);
	if(!baud)
		printf("Trying Baudrate %d...", rate[rate_cur]);
	legacy_with_timeout=1;
	timeout.tv_sec=1;				//1 sekunde beim ersten versuch, dann nurnoch die millisekunden...
	timeout.tv_usec=baud ? (safemode>5 ? 100000*safemode : 500000) : 100000;	//...und 100 ms timeout bei baudrate detection, ansonsten 500ms timeout
	do
	{
		uart_putc(term_fd, CMD_CONNECT);
		if(retval=uart_get_msg(term_fd, tempchar))
		{
			//printf("TIMEOUT\n");
			//printf("Writing connection request to bootloader..."); fflush(stdout);
			//uart_putc(term_fd, CMD_CONNECT);
			//if(retval=uart_get_msg(term_fd, tempchar))
			//{
			//	printf("TIMEOUT\n");
				if(!baud)
				{
					if((++rate_cur)>=rate_num)
						rate_cur=0;
					printf("\15Trying Baudrate %d...        ", rate[rate_cur]); fflush(stdout);
					deinit_terminal(term_fd);
					term_fd = init_terminal(MODEMDEVICE, get_rate(rate[rate_cur]));
				}
				else
					printf("."); fflush(stdout);
			//}
		}
		timeout.tv_sec=0;		//0 sekunden...nurnoch die millisekunden zhlen bei den folgeversuchen
	} while(retval);
	//puffer leeren (mit 0s timeout)
	uart_get_msg(term_fd, tempchar);
	timeout.tv_sec=2;		//2 sekunden...
	timeout.tv_usec=0;		//...und 0 ms timeout
	legacy_with_timeout=0;	//timeout aus
	
	proto=protokoll_supported(tempchar[0]<<8 | tempchar[1]);
	if(proto)												//Protokollversion prfen
		printf("\nConnection established...\nProtokoll v%d.%d OK...\n", tempchar[0], tempchar[1]);
	else
	{
		printf("\nError connecting\n");
		fprintf(stderr, "ERROR!!! - protokoll v%d.%d unknown...\n", tempchar[0], tempchar[1]);
		exit(2);
	}
	
	if(action_counter>0)
	{
		char *pos;
		
		printf("Parsing commands...\n");
		for(counter=0; counter<action_counter; counter++)
		{
			char *einleser=action[counter];
			char teil[3][8192];
			int teilcounter;
			
			for(teilcounter=0; teilcounter<3; teilcounter++)
			{
				pos=strstr(einleser, ":");
				if((pos==NULL) && (teilcounter<2))
					printf("General error parsing command %d, going on with next command...\n", counter);
				else
				{
					if(pos!=NULL)
						*pos='\0';
					strcpy(teil[teilcounter], einleser);
					if(pos!=NULL)
						*pos=':';
					einleser=pos+1;
				}
			}
			printf("\nExecuting parsed command '-U %s'...\n", action[counter]);
			if(strcmp(teil[0], "flash")==0)
			{
				if(strcmp(teil[1], "w")==0)
				{
					//Erase flash?
					if(erase_flash)
						do_erase_flash(term_fd);
					erase_flash=0;
					//write flash
					do_write_flash(term_fd, teil[2]);
				}
				else if(strcmp(teil[1], "r")==0)
				{
					erase_flash=0;
					//read flash
					do_read_flash(term_fd, teil[2]);
				}
				else
					printf("Error executing command %d [Selected target operation not supported (yet)], going on with next command...\n", counter+1);
			}
			else if(strcmp(teil[0], "eeprom")==0)
			{
				if(strcmp(teil[1], "w")==0)
				{
					//write eeprom
					do_write_eeprom(term_fd, teil[2]);
				}
				else if(strcmp(teil[1], "r")==0)
				{
					//read flash
					do_read_eeprom(term_fd, teil[2]);
				}
				else
					printf("Error executing command %d [Selected target operation not supported (yet)], going on with next command...\n", counter+1);
			}
			else if(strcmp(teil[0], "lfuse")==0)
			{
				if(strcmp(teil[1], "r")==0)
				{
					//read flash
					do_read_lfuse(term_fd);
				}
				else
					printf("Error executing command %d [Selected target operation not supported (yet)], going on with next command...\n", counter+1);
			}
			else if(strcmp(teil[0], "hfuse")==0)
			{
				if(strcmp(teil[1], "r")==0)
				{
					//read flash
					do_read_hfuse(term_fd);
				}
				else
					printf("Error executing command %d [Selected target operation not supported (yet)], going on with next command...\n", counter+1);
			}
			else
				printf("Error executing command %d [Selected target memory not supported (yet)], going on with next command...\n", counter+1);
		}
	}
	
	//Erase flash?
	if(erase_flash)
		do_erase_flash(term_fd);

	if(jump)
	{
		if(command_supported(CMD_JUMP_PROGRAM, proto))
		{
			printf("Waiting 2 seconds...\n");
			fflush(stdout);
			sleep(2);
			printf("Jumping to program...");
			fflush(stdout);
			uart_putc(term_fd, CMD_JUMP_PROGRAM);
			uart_get_msg(term_fd, tempchar);
			if(tempchar[0]==MSG_END)
				printf("OK\n");
			else
				printf("ERROR\n");
		}
		else
			printf("Error: Command CMD_JUMP_PROGRAM not supported...\n");
	}
	
	printf("\navr-programmer done. Thank you!\n\n");
	deinit_terminal(term_fd);
	return 0;
}

/*
atexit 
		deinit_terminal(term_fd);
			fclose(fihex);
*/
